@@ -1,14 +1,16 @@
 import asyncio
+import contextvars
 import copy
 import inspect
 import json
 import logging
 import os
+import signal
 import sys
 import traceback
 import uuid
 from functools import partial
-from typing import Optional, ClassVar, Awaitable
+from typing import Literal, Optional, ClassVar, Awaitable
 
 import requests
 from tornado import ioloop
@@ -17,9 +19,10 @@ from beaker_kernel.lib.config import reset_config, config
 from beaker_kernel.lib.context import BeakerContext, autodiscover_contexts
 from beaker_kernel.lib.jupyter_kernel_proxy import InterceptionFilter, JupyterMessage, KernelProxyManager
 from beaker_kernel.lib.utils import (message_handler, LogMessageEncoder, magic,
-                        handle_message, get_socket, execution_context, parent_message_context)
+                        handle_message, get_socket, execution_context, parent_message_context,
+                        ForwardMessage, ensure_async)
 
-USER_RESPONSE_WAIT_TIME = 100
+USER_RESPONSE_WAIT_TIME_SECONDS = 100
 
 logger = logging.getLogger(__name__)
 
@@ -67,6 +70,7 @@ class BeakerKernel(KernelProxyManager):
         self.internal_executions = set()
         self.subkernel_execution_tracking = {}
         self.running_actions = {}
+        context_args = session_config.get("context", {})
         super().__init__(session_config, session_id=f"{kernel_id}_session")
         self.register_magic_commands()
         self.add_base_intercepts()
@@ -74,13 +78,28 @@ class BeakerKernel(KernelProxyManager):
         self.user_responses = dict()
         # Initialize context (Using the event loop to simulate `await`ing the async func in non-async setup)
         event_loop = asyncio.get_event_loop()
-        context_task = event_loop.create_task(self.start_default_context())
+        context_task = event_loop.create_task(self.start_default_context(**context_args))
         context_task.add_done_callback(lambda task: None)
 
-    async def start_default_context(self):
-        default_context = os.environ.get('BEAKER_DEFAULT_CONTEXT')
-        default_context_payload = os.environ.get('BEAKER_DEFAULT_CONTEXT_PAYLOAD', "{}")
-        if default_context_payload:
+    async def start_default_context(self, default_context=None, default_context_payload=None, **options):
+        default_context = default_context or os.environ.get('BEAKER_DEFAULT_CONTEXT')
+        default_context_payload = default_context_payload or os.environ.get('BEAKER_DEFAULT_CONTEXT_PAYLOAD', "{}")
+
+        # Avoiding passing in optional args so defaults can be used
+        optional_args = {}
+        language = options.get("language", None) or os.environ.get('BEAKER_DEFAULT_CONTEXT_LANGUAGE', None)
+        if language:
+            optional_args["language"] = language
+
+        # Set context specific options
+        debug = options.get("debug", None) or os.environ.get('BEAKER_DEFAULT_CONTEXT_DEBUG', None)
+        verbose = options.get("verbose", None) or os.environ.get('BEAKER_DEFAULT_CONTEXT_VERBOSE', None)
+        if debug is not None:
+            self.debug_enabled = debug
+        if verbose is not None:
+            self.verbose = verbose
+
+        if isinstance(default_context_payload, str):
             try:
                 default_context_payload = json.loads(default_context_payload)
             except json.JSONDecodeError:
@@ -93,7 +112,7 @@ class BeakerKernel(KernelProxyManager):
         if not default_context:
             default_context = "default"
             default_context_payload = {}
-        await self.set_context(default_context, None)
+        await self.set_context(default_context, default_context_payload, **optional_args)
 
     def add_base_intercepts(self):
         """
@@ -117,6 +136,9 @@ class BeakerKernel(KernelProxyManager):
         self.server.intercept_message("shell", "set_agent_model", self.set_agent_model)
         self.server.intercept_message("shell", "reset_request", self.reset_kernel)
         self.server.intercept_message("control", "interrupt_request", self.interrupt)
+        self.server.intercept_message("control", "shutdown_request", self.shutdown)
+        self.server.intercept_message("shell", "notebook_state_response", self.notebook_state_response)
+        self.server.intercept_message("shell", "beaker_session_info_request", self.beaker_session_info)
 
     def register_magic_commands(self):
         for _, method in inspect.getmembers(self, lambda member: inspect.ismethod(member) and hasattr(member, "_magic_prefix")):
@@ -232,7 +254,9 @@ class BeakerKernel(KernelProxyManager):
             )
             return
         socket = get_socket(stream)
-        self.server.filters.remove(InterceptionFilter(socket, msg_type, func))
+        filter = InterceptionFilter(socket, msg_type, func)
+        if filter in self.server.filters:
+            self.server.filters.remove(filter)
 
     async def send_preview(self, parent_header=None):
         if self.context.preview:
@@ -241,6 +265,48 @@ class BeakerKernel(KernelProxyManager):
                 if preview_payload:
                     self.send_response("iopub", "preview", preview_payload, parent_header=parent_header)
 
+    async def send_kernel_state_info(self, parent_header=None):
+        if self.context.kernel_state:
+            with execution_context("kernel_state_info"):
+                state_payload = await self.context.kernel_state()
+                if state_payload:
+                    self.send_response("iopub", "kernel_state_info", state_payload, parent_header=parent_header)
+
+    async def send_chat_history(self, parent_header=None):
+        if self.context.agent.chat_history:
+            from archytas.chat_history import OutboundChatHistory, OutboundModel, SummaryRecord, MessageRecord
+            from dataclasses import asdict
+            chat_history = self.context.agent.chat_history
+            model = self.context.agent.model
+            records = await chat_history.records(auto_update_context=False)
+            output = OutboundChatHistory(
+                records=[{
+                    "message": {
+                        "text": record.message.text(),
+                        "raw_content": record.message.content,
+                        **record.message.model_dump(),
+                    },
+                    "uuid": record.uuid,
+                    "token_count": record.token_count,
+                    "metadata": record.metadata,
+                    "react_loop_id": record.react_loop_id,
+                } for record in records],
+                system_message=chat_history.system_message.message.text(),
+                tool_token_usage_estimate=chat_history.tool_token_estimate,
+                model=OutboundModel(
+                    provider=model.__class__.__name__,
+                    model_name=model.model_name,
+                    context_window=model.contextsize()
+                ),
+                message_token_count=sum(record.token_count for record in records if isinstance(record, MessageRecord) and record.token_count),
+                summary_token_count=sum(record.token_count for record in records if isinstance(record, SummaryRecord) and record.token_count),
+                overhead_token_count=chat_history.token_overhead,
+                summarization_threshold=model.summarization_threshold,
+                token_estimate=chat_history._token_estimate or await chat_history.token_estimate(model),
+            )
+            output_dict = asdict(output)
+            self.send_response("iopub", "chat_history", output_dict, parent_header=parent_header)
+
     async def update_connection_file(self, **kwargs):
         try:
             with open(self.connection_file, "r") as connection_file:
@@ -277,12 +343,10 @@ class BeakerKernel(KernelProxyManager):
         subkernel = self.context.subkernel
         kernel_setup_func = getattr(subkernel, "setup", None)
         with execution_context(type="setup", name=context_name, parent_header=parent_header):
-            if inspect.iscoroutinefunction(kernel_setup_func):
-                await kernel_setup_func()
-            elif inspect.isfunction(kernel_setup_func) or inspect.ismethod(kernel_setup_func):
-                kernel_setup_func()
+            await ensure_async(kernel_setup_func())
         await self.update_connection_file(context={"name": context_name, "config": context_info})
         await self.send_preview(parent_header=parent_header)
+        await self.send_kernel_state_info(parent_header=parent_header)
 
     async def post_execute(self, queue, message_id, data):
         message = JupyterMessage.parse(data)
@@ -304,12 +368,16 @@ class BeakerKernel(KernelProxyManager):
             This allows the normal execution flow to respond quickly in case these tasks are slow
             or resource intensive.
             """
+            coroutines = []
             if post_execute and (callable(post_execute) or inspect.iscoroutinefunction(post_execute)):
                 # If we have a callback function, then add it as a task to the execution loop so it runs
-                await post_execute(message)
+                coroutines.append(post_execute(message))
             # Always only generate and send preview after post_execute completes in case state changes or setup is
             # performed in the post_execute function
-            await self.send_preview(parent_header=message.parent_header)
+            coroutines.append(self.send_preview(parent_header=message.parent_header))
+            coroutines.append(self.send_kernel_state_info(parent_header=message.parent_header))
+            coroutines.append(self.send_chat_history(parent_header=message.parent_header))
+            await asyncio.gather(*coroutines)
 
         if loop:
             loop.create_task(task())
@@ -351,16 +419,25 @@ class BeakerKernel(KernelProxyManager):
 
     @message_handler
     async def interrupt(self, _message):
-        try:
-            subkernel_id = self.context.subkernel.jupyter_id
-            print(f"Interrupting connected subkernel: {subkernel_id}")
-            requests.post(
-                f"{self.context.beaker_kernel.jupyter_server}/api/kernels/{subkernel_id}/interrupt",
-                headers={"Authorization": f"token {config.jupyter_token}"},
-            )
-        except requests.exceptions.HTTPError as err:
-            logger.error(f"Subkernel cannot be interrupted.\nDetails:\n  {err.request.body}", exc_info=err)
-            return None
+        self._interrupt(interrupt_subkernel=True)
+
+    def soft_interrupt(self, signal, frame):
+        self._interrupt(interrupt_subkernel=False)
+
+    def _interrupt(self, interrupt_subkernel=True):
+        if interrupt_subkernel:
+            try:
+                subkernel_id = self.context.subkernel.jupyter_id
+                print(f"Interrupting connected subkernel: {subkernel_id}")
+                requests.post(
+                    f"{self.context.beaker_kernel.jupyter_server}/api/kernels/{subkernel_id}/interrupt",
+                    headers={"Authorization": f"token {config.jupyter_token}"},
+                    timeout=0.5,
+                )
+            except requests.exceptions.HTTPError as err:
+                logger.error(f"Subkernel cannot be interrupted.\nDetails:\n  {err.request.body}", exc_info=err)
+            except requests.exceptions.ConnectionError:
+                logger.error(f"Subkernel cannot be interrupted.\nDetails:\n  {err}", exc_info=err)
 
         for key, value in list(self.running_actions.items()):
             if inspect.iscoroutine(value):
@@ -370,22 +447,34 @@ class BeakerKernel(KernelProxyManager):
             del self.running_actions[key]
         return None
 
-    async def prompt_user(self, query, parent_message=None):
+    @message_handler
+    async def shutdown(self, message):
+        def stop_loop(loop: ioloop.IOLoop):
+            loop.stop()
+        self.context.cleanup()
+        self.context = None
+        # Stop loop after short delay to allow cleanup to run.
+        loop = ioloop.IOLoop.current()
+        loop.call_later(0.2, stop_loop, loop)
+        return None
+
+    async def prompt_user(self, query, parent_message=None, format: Optional[Literal['workflow_confirmation']]=None):
         msg_id = str(uuid.uuid4())
         self.send_response(
             "stdin",
             "input_request",
-            {"prompt": query},
+            {"prompt": query, "format": format},
             parent_header=getattr(parent_message, "header", None),
             parent_identities=getattr(parent_message, "identities", None),
             msg_id=msg_id,
         )
-        for _ in range(USER_RESPONSE_WAIT_TIME):
+        sleep_duration = 0.2
+        for _ in range(round(USER_RESPONSE_WAIT_TIME_SECONDS / sleep_duration)):
             if msg_id in self.user_responses:
                 result = self.user_responses[msg_id]
                 del self.user_responses[msg_id]
                 return result
-            await asyncio.sleep(1)
+            await asyncio.sleep(sleep_duration)
 
         raise Exception("Query timed out. User took too long to respond.")
 
@@ -434,16 +523,20 @@ class BeakerKernel(KernelProxyManager):
         self.send_stream_message(stream_type="stderr", text=text, parent_header=parent_header)
 
     @message_handler
-    async def llm_request(self, message):
-        # Send "code" to LLM Agent. The "code" is actually the LLM query
+    async def llm_request(self, message: JupyterMessage):
         from archytas.exceptions import AuthenticationError
-        content = message.content
+        content: dict = message.content
         request = content.get("request", None)
         if not request:
             return
         if not self.context:
             raise Exception("Context has not been set")
         setattr(self.context, "current_llm_query", request)
+
+        notebook_state = message.metadata.get("notebook_state", None)
+        if notebook_state is not None:
+            self.context.notebook_state = notebook_state
+
         request_key = f"llm_query:{message.header['msg_id']}"
         try:
             try:
@@ -478,10 +571,10 @@ class BeakerKernel(KernelProxyManager):
                 raise
             except Exception as err:
                 error_text = f"""LLM Error:
-{err}
+    {err}
 
-{traceback.format_exc()}
-"""
+    {traceback.format_exc()}
+    """
                 stream_content = {"name": "stderr", "text": error_text}
                 self.send_response(
                     "iopub", "stream", stream_content, parent_header=message.header
@@ -521,17 +614,26 @@ class BeakerKernel(KernelProxyManager):
         finally:
             if request_key in self.running_actions:
                 del self.running_actions[request_key]
+        await self.send_chat_history(message.header)
 
     @message_handler
     async def context_info_request(self, message):
-        context_slugs_by_class = dict((cls, slug) for slug, cls in AVAILABLE_CONTEXTS.items())
-        context_class = self.context.__class__
-        context_slug = context_slugs_by_class.get(context_class, "Not Found")
-        full_context_class = f"{context_class.__module__}.{context_class.__name__}"
-        context_config = getattr(self.context, "config", {}).get("context_info", None)
-        context_info = self.context.get_info()
-        language_slug = self.context.subkernel.SLUG
-        subkernel_name = self.context.subkernel.KERNEL_NAME
+        if self.context is not None:
+            context_slugs_by_class = dict((cls, slug) for slug, cls in AVAILABLE_CONTEXTS.items())
+            context_class = self.context.__class__
+            context_slug = context_slugs_by_class.get(context_class, "Not Found")
+            full_context_class = f"{context_class.__module__}.{context_class.__name__}"
+            context_config = getattr(self.context, "config", {}).get("context_info", None)
+            context_info = await self.context.get_info()
+            language_slug = self.context.subkernel.SLUG
+            subkernel_name = self.context.subkernel.KERNEL_NAME
+        else:
+            context_slug = "NONE"
+            full_context_class = "None"
+            context_config = None
+            language_slug = "Not set"
+            subkernel_name = "Not set"
+            context_info = None
 
         self.send_response(
             stream="iopub",
@@ -548,6 +650,7 @@ class BeakerKernel(KernelProxyManager):
             },
             parent_header=message.header,
         )
+        await self.send_chat_history(message.header)
         return None
 
     @message_handler
@@ -572,7 +675,7 @@ class BeakerKernel(KernelProxyManager):
             await self.set_context(context_name, context_info, language=language, parent_header=parent_header)
 
         # Send context_response
-        context_response_content = self.context.get_info()
+        context_response_content = await self.context.get_info()
         self.send_response(
             stream="iopub",
             msg_or_type="context_setup_response",
@@ -580,7 +683,7 @@ class BeakerKernel(KernelProxyManager):
             parent_header=parent_header,
         )
 
-    @message_handler
+    @message_handler(send_status_updates=False, send_reply=False)
     async def input_reply(self, message):
         content = message.content
         parent_id = message.parent_header["msg_id"]
@@ -598,6 +701,7 @@ class BeakerKernel(KernelProxyManager):
             model = config.get_model()
         if model:
             self.context.agent.model = model
+        await self.send_chat_history(message.header)
 
     @message_handler
     async def reset_kernel(self, message):
@@ -608,8 +712,40 @@ class BeakerKernel(KernelProxyManager):
             language=self.context.subkernel.SLUG,
             parent_header=message.header
         )
+        await self.send_chat_history(message.header)
         return True
 
+    @message_handler
+    async def beaker_session_info(self, message):
+        return await self.context.get_info()
+
+    async def notebook_state_response(self, server, target_stream, data):
+        async with handle_message(server, target_stream, data, send_status_updates=False, send_reply=False) as ctx:
+            setattr(self.notebook_state_response.__func__, 'result', ctx.message.content)
+            return None
+    setattr(notebook_state_response, 'result', None)
+
+    async def request_notebook_state(self, parent_message=None):
+        msg_id = str(uuid.uuid4())
+        self.send_response(
+            "iopub",
+            "notebook_state_request",
+            {},
+            parent_header=getattr(parent_message, "header", None),
+            parent_identities=getattr(parent_message, "identities", None),
+            msg_id=msg_id,
+        )
+        timeout = 1
+        steps = 10
+        sleep_duration = timeout / steps
+        for _ in range(steps):
+            result = getattr(self.notebook_state_response, 'result', None)
+            if result != None:
+                setattr(self.notebook_state_response.__func__, 'result', None)
+                return result
+            await asyncio.sleep(sleep_duration)
+        return None
+
 
 # Provided for backwards compatibility
 LLMKernel = BeakerKernel
@@ -617,11 +753,13 @@ LLMKernel = BeakerKernel
 
 def cleanup(kernel: BeakerKernel):
     try:
-        kernel.context.cleanup()
+        if kernel.context is not None:
+            kernel.context.cleanup()
     except requests.exceptions.ConnectionError:
         print("Unable to connect to server. Possible server shutdown.")
     except Exception as err:
-        print(f"Couldn't shutdown subkernel: {err}")
+        print(f"Error during cleanup: {err}")
+        traceback.print_exc()
 
 
 def start(connection_file):
@@ -635,10 +773,13 @@ def start(connection_file):
     kernel = BeakerKernel(notebook_config, kernel_id=kernel_id, connection_file=connection_file)
 
     try:
+        # Catch INTERRUPT signals and treat them as soft interrupts.
+        signal.signal(signal.SIGINT, kernel.soft_interrupt)
         loop.start()
-    except KeyboardInterrupt:
+    finally:
         # Perform shutdown cleanup here
         cleanup(kernel)
+        signal.signal(signal.SIGINT, signal.SIG_DFL)
         sys.exit(0)
 
 
