@@ -5,6 +5,7 @@ import importlib
 import inspect
 import logging
 import os
+import tempfile
 import toml
 from dataclasses import dataclass, field, MISSING, asdict, is_dataclass
 from enum import Enum
@@ -109,6 +110,7 @@ def configfield(
     placeholder_text: str = MISSING,
     label: str = MISSING,
     options: Callable[[], list[Any]] = MISSING,
+    **extra: dict[str, Any],
 ):
     def dynamic_factory():
         # Get value if defined
@@ -142,6 +144,7 @@ def configfield(
         "sensitive": sensitive,
         "aliases": aliases,
         "save_default_value": save_default_value,
+        "extra": extra,
     }
     if env_var and env_var is not MISSING:
         metadata["env_var"] = env_var
@@ -188,6 +191,13 @@ class LLM_Service_Provider:
         sensitive=True,
         save_default_value=True,
     )
+    summarization_threshold_pct: int = configfield(
+        description="Percentage of the total model's available context window to allow to be used prior to summarizing.",
+        default=20,
+        save_default_value=True,
+        min=0,
+        max=100,
+    )
 
     @classmethod
     def default_value(cls):
@@ -220,8 +230,34 @@ class ConfigClass:
         default=True,
         sensitive=False,
         normalize_function=normalize_bool,
-        label="Foo???"
+        label="Enable Checkpointing?"
     )
+    beaker_run_path: os.PathLike = configfield(
+        description="Path to use for beaker run items such as kernel json files and checkpoint data",
+        env_var="BEAKER_RUN_PATH",
+        default_factory=lambda: os.path.expanduser("~/.local/share/beaker/runtime"),
+        save_default_value=False,
+    )
+    send_notebook_state: bool = configfield(
+        description="Flag as to whether to include the state of the notebook on agent query execution.",
+        env_var="SEND_NOTEBOOK_STATE",
+        default=False,
+        sensitive=False,
+        normalize_function=normalize_bool,
+        label="Send notebook state on query?"
+    )
+    send_kernel_state: bool = configfield(
+        description="Flag as to whether to include the state of the subkernel on agent query execution.",
+        env_var="SEND_KERNEL_STATE",
+        default=True,
+        sensitive=False,
+        normalize_function=normalize_bool,
+        label="Send kernel state on query?"
+    )
+
+    @property
+    def checkpoint_storage_path(self):
+        return os.path.join(self.beaker_run_path, "checkpoints")
 
     tools_enabled: Table[bool] = configfield(
         description="This table allows you to enable/disable tools. The key is the name of the tool, and the value is a \
@@ -248,6 +284,11 @@ boolean value which will enable/disable the tool based on the value.",
                 "default_model_name": "claude-3-5-sonnet-20241022",
                 "api_key": ""
             },
+            "bedrock": {
+                "import_path": "archytas.models.bedrock.BedrockModel",
+                "default_model_name": "us.anthropic.claude-3-5-sonnet-20241022-v2:0",
+                "api_key": ""
+            },
             "gemini": {
                 "import_path": "archytas.models.gemini.GeminiModel",
                 "default_model_name": "gemini-1.5-pro",
@@ -289,7 +330,6 @@ boolean value which will enable/disable the tool based on the value.",
         "LLM_SERVICE_TOKEN",
         default="",
         sensitive=True,
-        aliases=["OPENAI_API_KEY"],
         save_default_value=False,
         placeholder_text="Set this value to override default.",
     )
@@ -366,6 +406,7 @@ class Config(ConfigClass):
         FILE = "file"
         SESSION = "session"
         SERVER = "server"
+        SINGLE_CONTEXT = "single"
         OTHER = "other"
 
     def __init__(self, **kwargs) -> None:
@@ -373,6 +414,11 @@ class Config(ConfigClass):
         config_type_str: str = kwargs.pop("config_type", os.environ.get("CONFIG_TYPE", "file"))
         self.config_type = getattr(self.ConfigTypes, config_type_str.upper(), self.ConfigTypes.FILE)
         self.defaults = kwargs
+        if self.config_type == self.ConfigTypes.SERVER:
+            if not os.path.isdir(self.checkpoint_storage_path):
+                os.makedirs(self.checkpoint_storage_path, exist_ok=True, mode=0o777)
+            if os.stat(self.checkpoint_storage_path).st_mode & 0o777 != 0o777:
+                os.chmod(self.checkpoint_storage_path, 0o777)
 
     def __getattr__(self, name: str):
         if name in (['__dataclass_fields__', '__dataclass_params__'] + list(ConfigClass.__dataclass_fields__.keys())):
@@ -384,12 +430,15 @@ class Config(ConfigClass):
             return getattr(self.config_obj, name)
         raise AttributeError
 
-    def get_model(self, provider_id=None, model_config=None):
+    def get_model(self, provider_id=None, model_config=None, **config_overrides):
         from archytas.exceptions import AuthenticationError
         config_obj: dict | None = None
 
         if provider_id:
             config_obj = self.providers.get(provider_id, None)
+            if config_obj:
+                # Copying config so it doesn't get changed globally if modified in the model instance
+                config_obj = deepcopy(config_obj)
             if isinstance(config_obj, dict) and isinstance(model_config, dict):
                 config_obj.update(model_config)
         elif isinstance(model_config, dict):
@@ -397,8 +446,11 @@ class Config(ConfigClass):
 
         # Fetch config if model_config is None or empty.
         if not config_obj:
-            # Get model from config
-            config_obj: dict = self.providers.get(self.provider, {})
+            # Get model from config,opying config so it doesn't get changed globally if modified in the model instance
+            config_obj: dict = deepcopy(self.providers.get(self.provider, {}))
+
+            # Update config with passed in overrides
+            config_obj.update(config_overrides)
 
             # Override defaults if set
             if self.model_provider_import_path:
@@ -412,8 +464,6 @@ class Config(ConfigClass):
         if not "import_path" in config_obj:
             return None
 
-        # Copying config so it doesn't get changed globally if modified in the model instance
-        config_obj = deepcopy(config_obj)
         if "model_name" not in config_obj and "default_model_name" in config_obj:
             config_obj["model_name"] = config_obj["default_model_name"]
 
